/*****************************************************************************
 * x11.c: misc x11 functions
 *****************************************************************************
 * $Id: x11.c,v 1.5 2003/07/02 12:28:41 alainjj Exp $
 *****************************************************************************
 * Copyright (C) 2001 Keuleu and Maupatz
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************
 *
 * Original code:
 *
 * misc x11 functions:  pixmap handling (incl. MIT SHMEM), event
 *                      tracking for the TV widget.
 *
 *  (c) 1998 Gerd Knorr <kraxel@cs.tu-berlin.de>
 *
 *****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "config.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/XawInit.h>
#include <X11/Xaw/Simple.h>
#include <X11/Shell.h>

#ifdef HAVE_MITSHM
# include <sys/ipc.h>
# include <sys/shm.h>
# include <X11/extensions/XShm.h>
#endif

#ifdef HAVE_LIBXXF86DGA
# include <X11/extensions/xf86dga.h>
#endif

#ifdef HAVE_XV
# include <X11/extensions/Xv.h>
# include <X11/extensions/Xvlib.h>
#endif

#include "grab.h"
#include "videodev.h"
#include "x11.h"
#include "channel.h"
#include "plugin.h"
#include "deinterlace.h"
#include "memcpy.h"
#include "fourcc.h"
#if defined(HAVE_DIVX4LINUX) || defined(HAVE_FFMPEG) || defined(HAVE_XVID)
#include "divx.h"
#endif

#define DISPLAY             XtDisplay
#define SCREEN              XtScreen
#define WINDOW              XtWindow
#define DEL_TIMER(proc)     XtRemoveTimeOut(proc)
#define ADD_TIMER(proc)     XtAppAddTimeOut(app_context,200,proc,NULL)

#ifdef HAVE_XV
# ifdef HAVE_MITSHM
#  define XVPUTIMAGE(dpy,xvp,dr,gc,xv,a,b,c,d,e,f,w,h)                 \
    if (have_shmem)                                                  \
    XvShmPutImage(dpy,xvp,dr,gc,xv,a,b,c,d,e,f,w,h,False);          \
    else                                                             \
    XvPutImage(dpy,xvp,dr,gc,xv,a,b,c,d,e,f,w,h)
# else
#  define XVPUTIMAGE(dpy,xvp,dr,gc,xv,a,b,c,d,e,f,w,h)                 \
        XvPutImage(dpy,xvp,dr,gc,xv,a,b,c,d,e,f,w,h)
# endif
  uint8_t tmp_buf0[576*768*2];
  uint8_t tmp_buf1[576*768*2];
  uint8_t *deinterlace_buffer[] = {tmp_buf0, tmp_buf1};
#endif

#ifdef HAVE_MITSHM
# define XPUTIMAGE(dpy,dr,gc,xi,a,b,c,d,w,h)                     \
    if (have_shmem)                                              \
    XShmPutImage(dpy,dr,gc,xi,a,b,c,d,w,h,True);                 \
    else                                                         \
    XPutImage   (dpy,dr,gc,xi,a,b,c,d,w,h)
#else
# define XPUTIMAGE(dpy,dr,gc,xi,a,b,c,d,w,h)                     \
    XPutImage   (dpy,dr,gc,xi,a,b,c,d,w,h)
#endif

/* ------------------------------------------------------------------------ */
extern PLUGIN *plugin;
//extern int decode;
extern char *video_base;
extern int direct_video;
extern int direct_xsize;
extern int maxheight[], maxwidth[];
extern int cur_norm;
extern int hwscan;
extern int cur_deinterlace;

static char *video_offset;

extern XtAppContext app_context;

static int display_bits = 0;
static int display_bytes = 0;
static int pixmap_bytes = 0;
static int have_shmem = 0;
int have_xv = 0;
int have_video_xv = 0;
static int x11_error = 0;

#ifdef HAVE_XV
extern int disable_xv;
extern XvPortID cmdline_xv_img_port;
static XvImage *xvimage = NULL;
static Atom xvattribute;
static XvImageFormatValues *xvfo = NULL;

static int xvid = -1;
static XvPortID xv_img_port = -1;
static XvAdaptorInfo *xvai = NULL;
static XvAttribute *xvat = NULL;

static int xvver, xvrel, xvreq, xvev, xverr;
static int xvadaptors;
static Atom xv_img_colorkey = -1;
#  ifdef HAVE_VIDEO_XV
static XvPortID xv_video_port = -1;
static Atom xv_video_encoding = -1;
static Atom xv_video_brightness = -1;
static Atom xv_video_contrast = -1;
static Atom xv_video_saturation = -1;
static Atom xv_video_hue = -1;
static Atom xv_video_volume = -1;
static Atom xv_video_mute = -1;
static Atom xv_video_freq = -1;
static Atom xv_video_colorkey = -1;
static GC xv_gc = NULL;
#  endif
int video_displayframe_xv (get_frame cb);
#endif
int video_displayframe_x11 (get_frame cb);

/* ------------------------------------------------------------------------ */
/* video overlay stuff                                                      */

int x11_native_format;
int x11_pixmap_format;
int swidth, sheight;            /* screen  */
int hw_scaling = 1;             /* Xvideo does the scaling */

extern Widget app_shell;
extern Widget tv;
static Widget video, video_parent;
 int wx, wy, wwidth, wheight, wmap;    /* window  */
 int ox, oy, owidth, oheight;    /* overlay */
static int maxx = 768, maxy = 576;

static set_overlay overlay_cb;
static XtIntervalId overlay_refresh;
static int did_refresh, oc_count;
static int visibility = VisibilityFullyObscured;
static int conf = 1, move = 1, overlay_on = 0;
static struct OVERLAY_CLIP oc[32];

static XImage *ximage = NULL;
static void *ximage_shm = NULL;        /* used for MIT shmem */
static GC gc;

static char *events[] = {
  "0", "1",
  "KeyPress",
  "KeyRelease",
  "ButtonPress",
  "ButtonRelease",
  "MotionNotify",
  "EnterNotify",
  "LeaveNotify",
  "FocusIn",
  "FocusOut",
  "KeymapNotify",
  "Expose",
  "GraphicsExpose",
  "NoExpose",
  "VisibilityNotify",
  "CreateNotify",
  "DestroyNotify",
  "UnmapNotify",
  "MapNotify",
  "MapRequest",
  "ReparentNotify",
  "ConfigureNotify",
  "ConfigureRequest",
  "GravityNotify",
  "ResizeRequest",
  "CirculateNotify",
  "CirculateRequest",
  "PropertyNotify",
  "SelectionClear",
  "SelectionRequest",
  "SelectionNotify",
  "ColormapNotify",
  "ClientMessage",
  "MappingNotify"
};

extern char *cmd;
extern void do_command_mult(char *);


/* ------------------------------------------------------------------------ */
int
x11_init (Display * dpy, Window window)
{
  XVisualInfo *info, template;
  XPixmapFormatValues *pf;
  int found, i, n;
#ifdef HAVE_XV
  int val, j;
  long int k;
#endif
  int format = 0;
  Screen *scr = DefaultScreenOfDisplay (dpy);
  Window root;
  XWindowAttributes wts;
  XvEncodingInfo *xv_encoding_info;
  int first_xvport;
  int last_xvport;


#ifdef HAVE_MITSHM
  if (XShmQueryExtension (dpy))
    {
      have_shmem = 1;
      if (hwscan)
        printf("MIT Shm extension available\n\n");
    }
#endif

#ifdef HAVE_XV
  if (!disable_xv)
    {
      if (Success !=
          XvQueryExtension (dpy, &xvver, &xvrel, &xvreq, &xvev, &xverr))
        {
          if (debug||hwscan)
            fprintf (stderr, "Server does'nt support Xvideo!\n");
        }
      else if (Success !=
               XvQueryAdaptors (dpy, DefaultRootWindow (dpy), &xvadaptors,
                                &xvai))
        {
          if (debug)
            fprintf (stderr, "XvQueryAdaptors failed!\n");
        }
      else
        {
          if (debug||hwscan)
            fprintf (stderr, "Number of Xv adaptors available: %d\n\n", xvadaptors);
              
          for (i = 0; i < xvadaptors; i++)
            {
              for (j = 0; j < xvai[i].num_formats; j++)
                {
                  if (debug)
                    fprintf (stderr, "Format %d = depth:%d id:%x\n", j,
                             xvai[i].formats[j].depth,
                             (int) xvai[i].formats[j].visual_id);
                }
                  




              if (hwscan)
                {
                  /* just print some info's about the Xvideo port */
                  n = fprintf(stderr, "port %ld-%ld",
                              xvai[i].base_id, xvai[i].base_id+xvai[i].num_ports-1);
                                                              
                  if ((xvai[i].type & XvInputMask) &&
                      (xvai[i].type & XvImageMask))
                    {
                      if (xvai[i].num_ports == 1)
                        fprintf(stderr, "%*s[ -xv_img_port %ld ]", 40-n, "", xvai[i].base_id);
                      else
                        fprintf(stderr, "%*s[ -xv_img_port %ld (to %ld)]", 40-n, "", xvai[i].base_id, xvai[i].base_id+xvai[i].num_ports-1);
                    }
                  fprintf(stderr, "\n");

                  fprintf(stderr, "    name : %s\n", xvai[i].name);

                  if ((xvai[i].type & XvInputMask) &&
                      (xvai[i].type & XvVideoMask))
                    {
                      fprintf(stderr, "    type : Xvideo, video overlay\n");
                      for (k=xvai[i].base_id; k<xvai[i].base_id+xvai[i].num_ports; k++)
                        {
                          XvQueryEncodings(dpy, k, &n, &xv_encoding_info);
                          fprintf(stderr, "        encodings supported on port %ld :\n", k);
                          for (j = 0; j < n; j++)
                            {
                              fprintf(stderr, "            %-20s %ldx%ld at %5d frames per %5d s\n",
                                      xv_encoding_info[j].name,
                                      xv_encoding_info[j].width, xv_encoding_info[j].height,
                                      xv_encoding_info[j].rate.denominator,
                                      xv_encoding_info[j].rate.numerator);
                            }
                          fprintf(stderr, "\n");
                          XvFreeEncodingInfo(xv_encoding_info);
                        }
                    }

                  if ((xvai[i].type & XvInputMask) &&
                      (xvai[i].type & XvImageMask))
                    {
                      fprintf(stderr, "    type : Xvideo, image scaler\n");
                      for (k=xvai[i].base_id; k<xvai[i].base_id+xvai[i].num_ports; k++)
                        {
                          xvfo = XvListImageFormats (dpy, k, &n);
                          fprintf(stderr, "        format supported on port %ld :", k);
                          for (j = 0; j < n; j++)
                            {
                              int t;
                              t = xvfo[j].id;
                              fprintf (stderr, " %c%c%c%c", t & 0xff,
                                      (t >> 8) & 0xff, (t >> 16) & 0xff,
                                      (t >> 24) & 0xff);
                              if (xvfo[j].type == XvRGB)
                                fprintf(stderr," [RGB %d, masks=0x%x/0x%x/0x%x]",
                                        xvfo[j].depth, xvfo[j].red_mask, xvfo[j].green_mask,
                                        xvfo[j].blue_mask);
                            }
                          fprintf(stderr, "\n");
                          if (xvfo != NULL)
                            XFree (xvfo);
                        }
                    }

                  fprintf(stderr, "\n");
                } // if (hwscan)





              /* selects FIRST Xv image port available which supports YUY2*/
              if ((xvai[i].type & XvInputMask) &&
                  (xvai[i].type & XvImageMask) && (xv_img_port == -1))
                {
                  if (cmdline_xv_img_port == 0)
                    { 
                      first_xvport = xvai[i].base_id;
                      last_xvport = xvai[i].base_id+xvai[i].num_ports-1;
                    }
                  else if (cmdline_xv_img_port >= xvai[i].base_id  &&
                           cmdline_xv_img_port <  xvai[i].base_id+xvai[i].num_ports)
                    {
                      first_xvport = last_xvport = cmdline_xv_img_port;
                    }
                  else
                    {
                      // set first and last values to odd values to be sure not to enter the for loop
                      first_xvport = 2;
                      last_xvport = 1;
                    }
                  for (k = first_xvport; k <= last_xvport; k++)
                    {
                      xvfo = XvListImageFormats (dpy, k, &n);
                      for (j = 0; j < n; j++)
                        {
                          if (xvfo[j].id == YUY2)
                            //if (xvfo[j].id == UYVY)
                            {
                              xv_img_port = k;
                              xvid = xvfo[j].id;
                              if (!hwscan)
                                {
                                  if (cmdline_xv_img_port == 0)
                                    fprintf (stderr, "Selected ");
                                  else
                                    fprintf (stderr, "Supplied ");
                                  fprintf (stderr, "XvImage adaptor with YUY2 support: %s on port %ld (grabdisplay)\n", xvai[i].name, xv_img_port);
                                }
                            }
                        }
                      if (xvfo != NULL)
                        XFree (xvfo);
                      
                      /*  Xv port is exclusive for xawdecode. */
                      if (XvGrabPort (dpy, xv_img_port, CurrentTime) != Success)
                        {
                          //fprintf (stderr, "XvGrabPort() failed.\n");
                          fprintf (stderr, "XvGrabPort() failed, XvImage port may be used by another application.\n");
                          xv_img_port = -1;
                        }
                      else
                        {
                          // Gotten a valid, unsued, Xv port
                          //XvUngrabPort (dpy, xv_img_port, CurrentTime);
                          break;
                        }
                    }
                }

#ifdef HAVE_VIDEO_XV
              /* selects FIRST Xv video port available */
              if ((xvai[i].type & XvInputMask) &&
                  (xvai[i].type & XvVideoMask) && (xv_video_port == -1))
                {
                  first_xvport = xvai[i].base_id;
                  last_xvport = xvai[i].base_id+xvai[i].num_ports-1;
                  for (k = first_xvport; k <= last_xvport; k++)
                    {
                      /*  Xv port is exclusive for xawdecode. */
                      if (XvGrabPort (dpy, k, CurrentTime) == Success)
                        {
                          xv_video_port = k;
                          fprintf (stderr, "Selected XvVideo: %s on port %ld (overlay)\n", xvai[i].name, xv_video_port);
                          XvUngrabPort (dpy, k, CurrentTime);
                          break;
                        }
                      else
                        fprintf (stderr, "XvGrabPort() failed, XvVideo port may be used by another application.\n");
                    }
                }
#endif // HAVE_VIDEO_XV

            }
        }


#ifdef HAVE_VIDEO_XV
      if (-1 == xv_video_port)
        {
          fprintf (stderr, "No XvVideo port available.\n");
          have_video_xv = 0;
        }
      else
        {
          have_video_xv = 1;
          xvat = XvQueryPortAttributes (dpy, xv_video_port, &n);
          if (xvat != NULL)
            {
              for (i = 0; i < n; i++)
                {
                  if (xvat[i].flags & XvSettable)
                    {
                      printf("%s\n", xvat[i].name);
                      if (strcmp ("XV_ENCODING", xvat[i].name) == 0)
                        {
                          xv_video_encoding = xvattribute = XInternAtom (dpy, xvat[i].name, True);
                          //val = xvat[i].min_value;
                          //XvSetPortAttribute (dpy, xv_video_port, xvattribute, val);
                          //if (debug)
                          //  printf("XV_ENCODING = %d\n", val);
                        }

                      if (strcmp ("XV_COLORKEY", xvat[i].name) == 0)
                        {
                          xv_video_colorkey = xvattribute = XInternAtom (dpy, xvat[i].name, True);
                        }

                      if (strcmp ("XV_BRIGHTNESS", xvat[i].name) == 0)
                        {
                          xv_video_brightness = xvattribute = XInternAtom (dpy, xvat[i].name, True);
                        }

                      if (strcmp ("XV_CONTRAST", xvat[i].name) == 0)
                        {
                          xv_video_contrast = xvattribute = XInternAtom (dpy, xvat[i].name, True);
                        }

                      if (strcmp ("XV_SATURATION", xvat[i].name) == 0)
                        {
                          xv_video_saturation = xvattribute = XInternAtom (dpy, xvat[i].name, True);
                        }

                      if (strcmp ("XV_HUE", xvat[i].name) == 0)
                        {
                          xv_video_hue = xvattribute = XInternAtom (dpy, xvat[i].name, True);
                        }

                      if (strcmp ("XV_FREQ", xvat[i].name) == 0)
                        {
                          xv_video_freq = xvattribute = XInternAtom (dpy, xvat[i].name, True);
                        }

                      if (strcmp ("XV_VOLUME", xvat[i].name) == 0)
                        {
                          xv_video_volume = xvattribute = XInternAtom (dpy, xvat[i].name, True);
                          fprintf(stderr, "XV_VOLUME max value = %d\n", xvat[i].max_value);
                          fprintf(stderr, "XV_VOLUME min value = %d\n", xvat[i].min_value);
                        }

                      if (strcmp ("XV_MUTE", xvat[i].name) == 0)
                        {
                          xv_video_mute = xvattribute = XInternAtom (dpy, xvat[i].name, True);
                          fprintf(stderr, "XV_MUTE max value = %d\n", xvat[i].max_value);
                          fprintf(stderr, "XV_MUTE min value = %d\n", xvat[i].min_value);
                        }

                      if (strcmp ("XV_SET_DEFAULTS", xvat[i].name) == 0)
                        {
                          xvattribute = XInternAtom (dpy, xvat[i].name, True);
                          //XvSetPortAttribute(dpy, xv_video_port, xvattribute, True);
                        }

                    }
                }
            }
        }
#endif // HAVE_VIDEO_XV

      if (-1 == xv_img_port)
        {
          fprintf (stderr, "No XvImage port available.\n");
          have_xv = 0;
        }
      else if (xvid == -1)
        {
          fprintf (stderr, "XvImage port available but none with YUY2 support.\n");
          have_xv = 0;
        }
      else
        {
	  int find_xvsetdefaults=0;
          have_xv = 1;
          if (debug)
            fprintf (stderr, "XvImage port %ld supports YUY2 format\n\n", xv_img_port);
          xvat = XvQueryPortAttributes (dpy, xv_img_port, &n);
          if (xvat != NULL)
            {
              /* Set by default */
              for (i = 0; i < n; i++)
	      {
	       if (strcmp ("XV_SET_DEFAULTS", xvat[i].name) == 0)
                        {
                          xvattribute = XInternAtom (dpy, xvat[i].name, True);
                          val = xvat[i].max_value;
                          XvSetPortAttribute (dpy, xv_img_port, xvattribute, val);
                          if (debug)
                            printf("Set XV to defaults values");
                        }
              }
              /* Set manually */
	      if (!find_xvsetdefaults)
              for (i = 0; i < n; i++)
                {
                  if (xvat[i].flags & XvSettable)
                    {
                      if (strcmp ("XV_COLORKEY", xvat[i].name) == 0)
                        {
                          xv_img_colorkey = xvattribute = XInternAtom (dpy, xvat[i].name, True);
                          val = 0; /*Assume the original value is near zero*/
                          XvSetPortAttribute (dpy, xv_img_port, xvattribute, val);
                          if (debug)
                            printf("XV_COLORKEY = %d\n", val);
                        }

                      if (strcmp ("XV_BRIGHTNESS", xvat[i].name) == 0)
                        {
                          xvattribute = XInternAtom (dpy, xvat[i].name, True);
                          val = (xvat[i].min_value + xvat[i].max_value) / 2;
                          XvSetPortAttribute (dpy, xv_img_port, xvattribute, val);
                          if (debug)
                            printf("XV_BRIGHTNESS = %d\n", val);
                        }

                      if (strcmp ("XV_CONTRAST", xvat[i].name) == 0)
                        {
                          xvattribute = XInternAtom (dpy, xvat[i].name, True);
                          val = (xvat[i].min_value + xvat[i].max_value) / 2;
                          XvSetPortAttribute (dpy, xv_img_port, xvattribute, val);
                          if (debug)
                            printf("XV_CONTRAST = %d\n", val);
                        }

                      if (strcmp ("XV_SATURATION", xvat[i].name) == 0)
                        {
                          xvattribute = XInternAtom (dpy, xvat[i].name, True);
                          val = (xvat[i].min_value + xvat[i].max_value) / 2;
                          XvSetPortAttribute (dpy, xv_img_port, xvattribute, val);
                          if (debug)
                            printf("XV_SATURATION = %d\n", val);
                        }

                      if (strcmp ("XV_HUE", xvat[i].name) == 0)
                        {
                          xvattribute = XInternAtom (dpy, xvat[i].name, True);
                          /* ATI -1000 to 1000; Nvidia 0..360 in degrees?*/
                          /*Fixme: assume 0 is good ?*/
                          val = 0;
                          XvSetPortAttribute (dpy, xv_img_port, xvattribute, val);
                          if (debug)
                            printf("XV_HUE = %d\n", val);
                        }
                    }
                }
            }
        }

      if (xvadaptors > 0)
        XvFreeAdaptorInfo (xvai);
    }
  else
#endif
    {
      have_xv = 0;
      have_video_xv = 0;
    }

  /* Now select correct video renderer */
  if (have_xv)
    {
      video_displayframe = video_displayframe_xv;
      if (!plugin)
        hw_scaling = 0;
    }
  else
    {
      video_displayframe = video_displayframe_x11;
    }

  /* Ask for visual type */
  template.screen = XDefaultScreen (dpy);
  template.visualid = XVisualIDFromVisual (DefaultVisualOfScreen (scr));
  info = XGetVisualInfo (dpy, VisualIDMask | VisualScreenMask, &template,
                         &found);

  root = DefaultRootWindow (dpy);
  XGetWindowAttributes (dpy, root, &wts);
  display_bits = wts.depth;
  display_bytes = (display_bits + 7) / 8;

  pf = XListPixmapFormats (dpy, &n);
  for (i = 0; i < n; i++)
    {
      if (pf[i].depth == display_bits)
        pixmap_bytes = pf[i].bits_per_pixel / 8;
    }

  if (debug)
    {
      fprintf (stderr, "x11: color depth: "
               "%d bits, %d bytes - pixmap: %d bytes\n",
               display_bits, display_bytes, pixmap_bytes);
      if (info->class == TrueColor)
        {
          fprintf (stderr, "x11: color masks: "
                   "red=0x%08lx green=0x%08lx blue=0x%08lx\n",
                   info->red_mask, info->green_mask, info->blue_mask);
        }

    }

  if (info->class == TrueColor)
    {
      switch (pixmap_bytes /* display_bytes */ )
        {
        case 2:
          format = (display_bits == 15) ? VIDEO_RGB15 : VIDEO_RGB16;
          break;
        case 3:
          format = VIDEO_RGB24;
          break;
        case 4:
          format = VIDEO_RGB32;
          break;
        }
      switch (pixmap_bytes)
        {
        case 2:
          x11_pixmap_format =
            (display_bits == 15) ? VIDEO_RGB15 : VIDEO_RGB16;
          break;
        case 3:
          x11_pixmap_format = VIDEO_RGB24;
          break;
        case 4:
          x11_pixmap_format = VIDEO_RGB32;
          break;
        }
    }
  if (info->class == StaticGray && info->depth == 8)
    {
      format = VIDEO_GRAY;
      x11_pixmap_format = VIDEO_GRAY;
    }
  if (0 == format)
    {
      if (info->class == PseudoColor && info->depth == 8)
        {
          fprintf (stderr,
                   "\n"
                   "8-bit Pseudocolor Visual (256 colors) is *not* supported.\n"
                   "You can startup X11 either with 15 bpp (or more)...\n"
                   "    xinit -- -bpp 16\n"
                   "... or with StaticGray visual:\n"
                   "    xinit -- -cc StaticGray\n");
        }
      else
        {
          fprintf (stderr, "Sorry, I can't handle your strange display\n");
        }
      exit (1);
    }

#ifdef HAVE_XV
  if ((!disable_xv) && (have_xv))
    {
      if (debug)
        fprintf(stderr, "Forcing x11_pixmap_format to VIDEO_YUY2\n");
      x11_pixmap_format = VIDEO_YUY2;
    }
#endif

  XFree (info);
  return format;
}

int
x11_error_dev_null (Display * dpy, XErrorEvent * event)
{
  x11_error++;
  if (debug > 1)
    fprintf (stderr, " x11-error");
  return 0;
}

void
set_colorkey(Display * dpy, int val)
{
#ifdef HAVE_XV
  if (have_xv && (xv_img_colorkey != -1))
    {
      XvSetPortAttribute (dpy, xv_img_port, xv_img_colorkey, val);
      XvGetPortAttribute (dpy, xv_img_port, xv_img_colorkey, &val);
      if (debug)
        printf("x11: set_colorkey: XV_COLORKEY = %d\n", val);
    }
#endif
}

int
get_colorkey(Display * dpy)
{
  int val;

#ifdef HAVE_XV
  if (have_xv && (xv_img_colorkey != -1))
    {
      XvGetPortAttribute (dpy, xv_img_port, xv_img_colorkey, &val);
      if (debug)
        printf("x11; get_colorkey: XV_COLORKEY = %d\n", val);
    }
  else
#endif
    val = -1;

    return val;
}
/* ------------------------------------------------------------------------ */
/*     xvimage,ximage handling for grab & display                           */


/*******************************Xvimage**********************************/
#ifdef HAVE_XV
XvImage *
x11_create_xvimage (Display * dpy, int width, int height, void **shm)
{
  XvImage *xvimage = NULL;
  unsigned char *xvimage_data;
#ifdef HAVE_MITSHM
  XShmSegmentInfo *shminfo = NULL;
  void *old_handler = NULL;
#endif

  if (debug)
    fprintf (stderr, "video: x11_create_xvimage\n");
#ifdef HAVE_MITSHM
  if (have_shmem)
    {
      x11_error = 0;
      old_handler = XSetErrorHandler (x11_error_dev_null);
      (*shm) = shminfo = malloc (sizeof (XShmSegmentInfo));
      memset (shminfo, 0, sizeof (XShmSegmentInfo));
      xvimage =
        XvShmCreateImage (dpy, xv_img_port, xvid, NULL, width, height, shminfo);
      shminfo->shmid =
        shmget (IPC_PRIVATE, xvimage->data_size, IPC_CREAT | 0777);
      if (-1 == shminfo->shmid)
        {
          perror ("shmget");
          goto oom;
        }
      shminfo->shmaddr = (char *) shmat (shminfo->shmid, 0, 0);
      if ((void *) -1 == shminfo->shmaddr)
        {
          perror ("shmat");
          goto oom;
        }
      xvimage->data = shminfo->shmaddr;
      shminfo->readOnly = False;

      XShmAttach (dpy, shminfo);
      XSync (dpy, False);
      shmctl (shminfo->shmid, IPC_RMID, 0);

      if (x11_error)
        {
          have_shmem = 0;
          shmdt (shminfo->shmaddr);
          free (shminfo);
          shminfo = *shm = NULL;
          XFree (xvimage);
          xvimage = NULL;
        }

    }
  else
    {
      have_shmem = 0;
      free (shminfo);
      shminfo = *shm = NULL;
    }
  XSetErrorHandler (old_handler);
#endif

  if (xvimage == NULL)
    {
      (*shm) = NULL;
      if (NULL == (xvimage_data = malloc (width * height * 2)))
        {
          fprintf (stderr, "out of memory for create xvimage\n");
          goto oom;
        }
      xvimage =
        XvCreateImage (dpy, xv_img_port, xvid, xvimage_data, width, height);
    }
  memset (xvimage->data, 16, xvimage->data_size);

  return xvimage;

oom:;
#ifdef HAVE_MITSHM
  if (shminfo)
    {
      if (shminfo->shmid && shminfo->shmid != -1)
        shmctl (shminfo->shmid, IPC_RMID, 0);
      free (shminfo);
    }
#endif
  if (xvimage)
    XFree (xvimage);
  return NULL;
}

void
x11_destroy_xvimage (Display * dpy, XvImage * xvimage, void *shm)
{

#ifdef HAVE_MITSHM
  XShmSegmentInfo *shminfo = shm;

  if (shminfo)
    {
      XShmDetach (dpy, shminfo);
      shmdt (shminfo->shmaddr);
      shmctl (shminfo->shmid, IPC_RMID, 0);
      free (shminfo);
      XFree (xvimage);
    }
  else
    {
#endif
      if (xvimage->data != NULL)
        XFree (xvimage->data);
      XFree (xvimage);
#ifdef HAVE_MITSHM
    }
#endif
  if (debug)
    fprintf (stderr, "video: x11_destroy_xvimage\n");
}

#endif



/********************************XLIB *************************************/

XImage *
x11_create_ximage (Display * dpy, int width, int height, void **shm)
{
  XImage *ximage = NULL;
  unsigned char *ximage_data;
  Screen *scr = DefaultScreenOfDisplay (dpy);
#ifdef HAVE_MITSHM
  XShmSegmentInfo *shminfo = NULL;
  void *old_handler;
#endif

  if (debug)
    fprintf (stderr, "video: x11_create_ximage\n");
#ifdef HAVE_MITSHM
  if (have_shmem)
    {
      x11_error = 0;
      old_handler = XSetErrorHandler (x11_error_dev_null);
      (*shm) = shminfo = malloc (sizeof (XShmSegmentInfo));
      memset (shminfo, 0, sizeof (XShmSegmentInfo));
      ximage = XShmCreateImage (dpy,
                                DefaultVisualOfScreen (scr),
                                DefaultDepthOfScreen (scr),
                                ZPixmap, NULL, shminfo, width, height);
      if (ximage)
        {
          shminfo->shmid = shmget (IPC_PRIVATE,
                                   ximage->bytes_per_line * ximage->height,
                                   IPC_CREAT | 0777);
          if (-1 == shminfo->shmid)
            {
              perror ("shmget");
              goto oom;
            }
          shminfo->shmaddr = (char *) shmat (shminfo->shmid, 0, 0);
          if ((void *) -1 == shminfo->shmaddr)
            {
              perror ("shmat");
              goto oom;
            }
          ximage->data = shminfo->shmaddr;
          shminfo->readOnly = False;

          XShmAttach (dpy, shminfo);
          XSync (dpy, False);
          shmctl (shminfo->shmid, IPC_RMID, 0);
          if (x11_error)
            {
              have_shmem = 0;
              shmdt (shminfo->shmaddr);
              free (shminfo);
              shminfo = *shm = NULL;
              XDestroyImage (ximage);
              ximage = NULL;
            }
        }
      else
        {
          have_shmem = 0;
          free (shminfo);
          shminfo = *shm = NULL;
        }
      XSetErrorHandler (old_handler);
    }
#endif

  if (ximage == NULL)
    {
      (*shm) = NULL;
      if (NULL == (ximage_data = malloc (width * height * pixmap_bytes)))
        {
          fprintf (stderr, "out of memory\n");
          goto oom;
        }
      ximage = XCreateImage (dpy,
                             DefaultVisualOfScreen (scr),
                             DefaultDepthOfScreen (scr),
                             ZPixmap, 0, ximage_data, width, height, 8, 0);
    }
  memset (ximage->data, 0, ximage->bytes_per_line * ximage->height);
  return ximage;

oom:
#ifdef HAVE_MITSHM
  if (shminfo)
    {
      if (shminfo->shmid && shminfo->shmid != -1)
        shmctl (shminfo->shmid, IPC_RMID, 0);
      free (shminfo);
    }
#endif
  if (ximage)
    XDestroyImage (ximage);
  return NULL;
}

void
x11_destroy_ximage (Display * dpy, XImage * ximage, void *shm)
{
#ifdef HAVE_MITSHM
  XShmSegmentInfo *shminfo = shm;

  if (shminfo)
    {
      XShmDetach (dpy, shminfo);
      XDestroyImage (ximage);
      shmdt (shminfo->shmaddr);
      free (shminfo);
    }
  else
#endif
    XDestroyImage (ximage);

  if (debug)
    fprintf (stderr, "video: x11_destroy_ximage\n");
}

Pixmap
x11_create_pixmap (Widget w, unsigned char *byte_data,
                   int width, int height, char *label)
{
  static XFontStruct *font;
  static XColor color, dummy;

  Pixmap pixmap;
  XImage *pix_ximage;
  XGCValues values;
  GC gc;
  void *shm;
  Display *dpy = XtDisplay (w);
  Screen *scr = XtScreen (w);

  if (!font)
    {
      font = XLoadQueryFont (dpy, "fixed");
      XAllocNamedColor (dpy, DefaultColormapOfScreen (scr),
                        "yellow", &color, &dummy);
    }

  pixmap = XCreatePixmap (dpy,
                          RootWindowOfScreen (scr),
                          width, height, DefaultDepthOfScreen (scr));

  values.font = font->fid;
  values.foreground = color.pixel;
  gc = XCreateGC (dpy, pixmap, GCFont | GCForeground, &values);

  if ((pix_ximage = x11_create_ximage (dpy, width, height, &shm)) == NULL)
    {
      fprintf(stderr, "x11_create_pixmap: x11_create_ximage failed\n");
      XFreePixmap (dpy, pixmap);
      XFreeGC (dpy, gc);
      return 0;
    }
  
  memcpy (pix_ximage->data, byte_data, width * height * pixmap_bytes);

  XPUTIMAGE (dpy, pixmap, gc, pix_ximage, 0, 0, 0, 0, width, height);

  if (label)
    XDrawString (dpy, pixmap, gc, 5, height - 5, label, strlen (label));

  x11_destroy_ximage (dpy, pix_ximage, shm);

  XFreeGC (dpy, gc);
  return pixmap;
}


/* ------------------------------------------------------------------------ */

static void
add_clip (int x1, int y1, int x2, int y2)
{
  if (oc[oc_count].x1 != x1 || oc[oc_count].y1 != y1 ||
      oc[oc_count].x2 != x2 || oc[oc_count].y2 != y2)
    {
      conf = 1;
    }
  oc[oc_count].x1 = x1;
  oc[oc_count].y1 = y1;
  oc[oc_count].x2 = x2;
  oc[oc_count].y2 = y2;
  oc_count++;
}

static void
get_clips ()
{
  int x1, y1, x2, y2, lastcount;
  Display *dpy;
  XWindowAttributes wts;
  Window root, me, rroot, parent, *children;
  uint nchildren, i;
  void *old_handler = XSetErrorHandler (x11_error_dev_null);

  if (debug > 1)
    fprintf (stderr, " getclips");
  lastcount = oc_count;
  oc_count = 0;
  dpy = DISPLAY (video);

  if (ox < 0)
    add_clip (0, 0, (uint) (-ox), oheight);
  if (oy < 0)
    add_clip (0, 0, owidth, (uint) (-oy));
  if ((ox + owidth) > swidth)
    add_clip (swidth - ox, 0, owidth, oheight);
  if ((oy + oheight) > sheight)
    add_clip (0, sheight - oy, owidth, oheight);

  root = DefaultRootWindow (dpy);
  me = WINDOW (video);
  for (;;)
    {
      XQueryTree (dpy, me, &rroot, &parent, &children, &nchildren);
      XFree ((char *) children);
      /* fprintf(stderr,"me=0x%x, parent=0x%x\n",me,parent); */
      if (root == parent)
        break;
      me = parent;
    }
  XQueryTree (dpy, root, &rroot, &parent, &children, &nchildren);

  for (i = 0; i < nchildren; i++)
    if (children[i] == me)
      break;

  for (i++; i < nchildren; i++)
    {
      XGetWindowAttributes (dpy, children[i], &wts);
      if (!(wts.map_state & IsViewable))
        continue;

      x1 = wts.x - ox;
      y1 = wts.y - oy;
      x2 = x1 + wts.width + 2 * wts.border_width;
      y2 = y1 + wts.height + 2 * wts.border_width;
      if ((x2 < 0) || (x1 > (int) owidth) || (y2 < 0) || (y1 > (int) oheight))
        continue;

      if (x1 < 0)
        x1 = 0;
      if (y1 < 0)
        y1 = 0;
      if (x2 > (int) owidth)
        x2 = owidth;
      if (y2 > (int) oheight)
        y2 = oheight;
      add_clip (x1, y1, x2, y2);
    }
  XFree ((char *) children);

  if (lastcount != oc_count)
    conf = 1;
  XSetErrorHandler (old_handler);
}

static void
refresh_timer (XtPointer clientData, XtIntervalId * id)
{
  Window win = RootWindowOfScreen (SCREEN (video));
  Display *dpy = DISPLAY (video);
  XSetWindowAttributes xswa;
  unsigned long mask;
  Window tmp;

  if (!move && wmap && visibility == VisibilityUnobscured)
    {
      if (debug > 1)
        fprintf (stderr, "video: refresh skipped\n");
      return;
    }

  if (debug > 1)
    fprintf (stderr, "video: refresh\n");
  overlay_refresh = 0;
  if (wmap && visibility != VisibilityFullyObscured)
    did_refresh = 1;

  xswa.override_redirect = True;
  xswa.backing_store = NotUseful;
  xswa.save_under = False;
  mask = (CWSaveUnder | CWBackingStore | CWOverrideRedirect);
  tmp = XCreateWindow (dpy, win, 0, 0, swidth, sheight, 0,
                       CopyFromParent, InputOutput, CopyFromParent,
                       mask, &xswa);
  XMapWindow (dpy, tmp);
  XUnmapWindow (dpy, tmp);
  XDestroyWindow (dpy, tmp);
  move = 0;
}

static void
configure_overlay ()
{
#ifdef HAVE_XV
  if (have_video_xv)
    return;
#endif
  if (!overlay_cb)
    return;

  if (debug > 1)
    fprintf (stderr, "video: configure");
  if (wmap && visibility != VisibilityFullyObscured)
    {
      if (visibility == VisibilityPartiallyObscured)
        get_clips ();
      else
        oc_count = 0;

      if (debug > 1)
        fprintf (stderr, " %s\n", conf ? "yes" : "no");
      if (conf)
        {
          overlay_on = 1;
          overlay_cb (ox, oy, owidth, oheight,
                      x11_native_format, oc, oc_count);
          if (overlay_refresh)
            DEL_TIMER (overlay_refresh);
          overlay_refresh = ADD_TIMER (refresh_timer);
          conf = 0;
        }
    }
  else
    {
      if (debug > 1)
        fprintf (stderr, " off\n");
      if (conf && overlay_on)
        {
          overlay_on = 0;
          overlay_cb (0, 0, 0, 0, 0, NULL, 0);
          if (overlay_refresh)
            DEL_TIMER (overlay_refresh);
          overlay_refresh = ADD_TIMER (refresh_timer);
          conf = 0;
        }
    }
}


void
video_set_size(int w, int h, int force)
{
  XtVaSetValues (video_parent, XtNwidth, w, XtNheight, h, NULL);
  video_new_size(force);
}

void
video_new_size (int force)
{
  Dimension x, y, w, h;

  XtVaGetValues (video_parent, XtNx, &x, XtNy, &y,
                 XtNwidth, &w, XtNheight, &h, NULL);

  wx = x;
  if (wx > 32768)
    wx -= 65536;

  wy = y;
  if (wy > 32768)
    wy -= 65536;

  wwidth = w;
  if (wwidth > 32768)
    wwidth -= 65536;

  wheight = h;
  if (wheight > 32768)
    wheight -= 65536;

  if (debug > 1)
    fprintf (stderr, "video: shell: size %dx%d+%d+%d\n",
             wwidth, wheight, wx, wy);

  if (wwidth > maxx)
    owidth = maxx;
  else
    owidth = wwidth;

  if (wheight > maxy)
    oheight = maxy;
  else
    oheight = wheight;

  if (!force)
    {
      owidth &= ~3;
      oheight &= ~3;
      ox = wx + (wwidth - owidth) / 2;
      oy = wy + (wheight - oheight) / 2;
      ox &= ~3;

      if (ox < wx)
        ox += 4;

      if (ox + owidth > wx + wwidth)
        owidth -= 4;
    }

  if (ximage)
    {
      x11_destroy_ximage (DISPLAY (video), ximage, ximage_shm);
      ximage = NULL;
    }

#ifdef HAVE_XV
  if (xvimage)
    {
      x11_destroy_xvimage (DISPLAY (video), xvimage, ximage_shm);
      xvimage = NULL;
    }

#ifdef HAVE_VIDEO_XV
  if (have_video_xv && (xv_gc != NULL)&&(xv_video_port != -1))
    {
      int sx, sy, dx, dy;
      int sw, sh;
      int dw, dh;
      sx = sy = dx = dy = 0;
      sw = 768;
      sh = 576;
      dw = wwidth;
      dh = wheight;
      if (Success != XvPutVideo(DISPLAY(video), xv_video_port, WINDOW(video), xv_gc,
                                sx,sy,sw,sh, dx,dy,dw,dh))
        fprintf(stderr, "Xv video_overlay: XvPutVideo failed!\n");
      
      //if (debug)
      fprintf(stderr,"Xvideo: video: win=0x%lx, "
              "src=%dx%d+%d+%d dst=%dx%d+%d+%d\n",
              WINDOW(video), sw,sh,sx,sy, dw,dh,dx,dy);
    }
#endif // HAVE_VIDEO_XV

#endif // HAVE_XV

#ifdef HAVE_LIBXXF86VM
  if (direct_video)
    {
      int vp_width, vp_height;

      XF86DGAGetViewPortSize (DISPLAY (video),
                              XDefaultScreen (DISPLAY (video)), &vp_width,
                              &vp_height);
      video_offset =
        video_base + (vp_height - oheight) * direct_xsize + vp_width - owidth;
    }
#endif

  conf = 1;
  move = 1;
  configure_overlay ();
}

/* ------------------------------------------------------------------------ */

static void
video_event (Widget widget, XtPointer client_data, XEvent * event,
             Boolean * d)
{
  //static unsigned long int i=0;
  //i++;
  //fprintf(stderr, "x11: video_event: got an event %ld\n", i);
  if (widget == video_parent)
    {
      /* shell widget */
      switch (event->type)
        {
        case ConfigureNotify:
#if 0
          wx = event->xconfigure.x;
          wy = event->xconfigure.y;
          wwidth = event->xconfigure.width;
          wheight = event->xconfigure.height;
          if (debug > 1)
            fprintf (stderr, "video: shell: cfg %dx%d+%d+%d\n",
                     wwidth, wheight, wx, wy);
#endif
          video_new_size (0);
          break;
        case MapNotify:
          if (debug > 1)
            fprintf (stderr, "video: shell: map\n");
          wmap = 1;
          conf = 1;
          configure_overlay ();
          break;
        case UnmapNotify:
          if (debug > 1)
            fprintf (stderr, "video: shell: unmap\n");
          wmap = 0;
          conf = 1;
          configure_overlay ();
          break;
        default:
          if (debug > 1)
            fprintf (stderr, "video: shell: %s\n", events[event->type]);
        }
      return;

    }
  else
    {
      /* TV widget (+root window) */
      switch (event->type)
        {
        case Expose:
          if (event->xvisibility.window == WINDOW (video))
            {
              /* tv */
              if (!event->xexpose.count)
                {
                  if (did_refresh)
                    {
                      did_refresh = 0;
                      if (debug > 1)
                        fprintf (stderr, "video: tv: last refresh expose\n");
                    }
                  else
                    {
                      if (debug > 1)
                        fprintf (stderr, "video: tv: expose\n");
                      conf = 1;
                      configure_overlay ();
                    }
                }
            }
          break;
        case VisibilityNotify:
          if (event->xvisibility.window == WINDOW (video))
            {
              /* tv */
              visibility = event->xvisibility.state;
              if (debug > 1)
                fprintf (stderr, "video: tv: visibility %d%s\n",
                         event->xvisibility.state,
                         did_refresh ? " (ignored)" : "");
              if (did_refresh)
                {
                  if (event->xvisibility.state != VisibilityFullyObscured)
                    did_refresh = 0;
                }
              else
                {
                  conf = 1;
                  configure_overlay ();
                }
            }
          else
            {
              /* root */
              if (debug > 1)
                fprintf (stderr, "video: root: visibility\n");
            }
          break;
        case MapNotify:
        case UnmapNotify:
        case ConfigureNotify:
          if (event->xvisibility.window != WINDOW (video))
            {
              if (debug > 1)
                fprintf (stderr, "video: root: %s%s\n",
                         events[event->type],
                         did_refresh ? " (ignored)" : "");
              if (!did_refresh)
                configure_overlay ();
            }
          break;
        default:
          if (debug > 1)
            fprintf (stderr, "video: tv(+root): %s\n", events[event->type]);
          break;
        }
    }
}

void
video_setmax (int x, int y)
{
  int ow, oh;

  if (debug)
    fprintf (stderr, "video: new maxsize %dx%d\n", x, y);

  ow = owidth;
  oh = oheight;
  maxx = x;
  maxy = y;

  video_new_size (0);
}

void
video_overlay (set_overlay cb)
{
#ifdef HAVE_VIDEO_XV
  static int grabbed = 0;
  int sx,sy,dx,dy;
  int sw,sh;
  int dw, dh;

  if (have_video_xv)
    {
      if (cb)
        {
          sw = 768/*owidth*/;
          sh = 576/*oheight*/;
          dw = wwidth;
          dh = wheight;
          sx = sy = dx = dy = 0;
          if (NULL == xv_gc)
            xv_gc = XCreateGC(DISPLAY(video), WINDOW(video), 0, NULL);

          if (!grabbed)
            {
              if (Success == XvGrabPort(DISPLAY(video), xv_video_port, CurrentTime))
                grabbed = 1;
              else
                fprintf(stderr, "Xv video_overlay: XvGrabPort failed!\n");
            }

          if (grabbed)
            {
              if (Success != XvPutVideo(DISPLAY(video), xv_video_port, WINDOW(video), xv_gc,
                                        sx,sy,sw,sh, dx,dy,dw,dh))
                fprintf(stderr, "Xv video_overlay: XvPutVideo failed!\n");
              
              //if (debug)
                fprintf(stderr,"Xvideo: video: win=0x%lx, "
                        "src=%dx%d+%d+%d dst=%dx%d+%d+%d\n",
                        WINDOW(video), sw,sh,sx,sy, dw,dh,dx,dy);
            }
          else
            {
              fprintf(stderr,"Xvideo: port %ld busy\n", xv_video_port);
            }
        }
      else
        {
          if (grabbed)
            {
              XClearArea(DISPLAY(video), WINDOW(video), 0, 0, 0, 0, False);
              XvStopVideo(DISPLAY(video), xv_video_port, WINDOW(video));
              XvUngrabPort(DISPLAY(video), xv_video_port, CurrentTime);
              grabbed = 0;
              //if (debug)
                fprintf(stderr,"Xvideo: video off\n");
            }
        }
    }
  else
#endif // HAVE_VIDEO_XV
    {
      if (cb)
        {
          overlay_cb = cb;
          conf = 1;
          configure_overlay ();
        }
      else
        {
          if (overlay_cb)
            {
              overlay_on = 0;
              overlay_cb (0, 0, 0, 0, 0, NULL, 0);
              overlay_refresh = ADD_TIMER (refresh_timer);
            }
          overlay_cb = NULL;
        }
    }
}

int
video_displayframe_x11 (get_frame cb)
{
  static void *image_data;
  int ydbar;
  int ysbar;
  int src_w, src_h;
  // owidth and oheight are the size of the overlay (i.e. the TV image) (max value is 768x576)
  // wwidth and wheight are the size of the window whatever size it is

  if (!WINDOW (video))
    return -1;

  if (!gc)
    gc = XCreateGC (DISPLAY (video), WINDOW (video), 0, NULL);

  src_w = owidth;
  src_h = oheight;

  if (!ximage)
    {
      ximage =
        x11_create_ximage (DISPLAY (video), owidth, oheight, &ximage_shm);
      if (NULL == ximage)
        {
          fprintf (stderr, "oops: can't create ximage\n");
          exit (1);
        }
      XClearArea (DISPLAY (video), WINDOW (video), 0, 0, wwidth, wheight,
                  False);
    }

  if (direct_video)
    {
      if (NULL == cb (video_offset, owidth, oheight, 0, direct_xsize))
        return -1;
    }
  else
    {
      image_data = ximage->data;
      if (NULL == cb (image_data, owidth, /*oheight*/src_h, 0, owidth))
        return -1;
      
      if (plugin)
        {
          ysbar = plugin->get_ybar();

          ydbar = (int)(ysbar * (double)(wheight) / (double)maxheight[cur_norm] );

          if (!hw_scaling)
            ysbar = ydbar; // if sw scaling, then image is already scaled as the black border...
        }
      else
        {
          ysbar = 0;
          ydbar = 0;
        }

#if defined(HAVE_DIVX4LINUX) || defined(HAVE_FFMPEG) || defined(HAVE_XVID)
      if (is_divx_in_progress())
        {
          divx_encode(image_data+ydbar*owidth*pixmap_bytes);

          if (divx_display_frames() == 0)
            return 0;
        }
#endif

      if (ydbar > 0)
        {
          //XClearArea (DISPLAY(video), WINDOW(video), 0, 0, owidth, ydbar, False);
          //XClearArea (DISPLAY(video), WINDOW(video), 0, oheight-ydbar, owidth, ydbar, False);
          XClearArea (DISPLAY(video), WINDOW(video), ox-wx, oy-wy, owidth, ydbar, False);
          XClearArea (DISPLAY(video), WINDOW(video), ox-wx, wheight-ydbar-oy+wy, owidth, ydbar, False);
        }
      XPUTIMAGE (DISPLAY (video), WINDOW (video), gc, ximage,
                 0, 0 + ydbar,
                 ox - wx, oy - wy + ydbar,
                 owidth, oheight-2*ydbar);
    }
  if(cmd) {do_command_mult(cmd);cmd=NULL;}
  return 0;
}

#ifdef HAVE_XV
int
video_displayframe_xv (get_frame cb)
{
  static void *image_data;
  int ydbar;
  int ysbar;
  int src_w, src_h;
  // owidth and oheight are the size of the overlay (i.e. the TV image) (max value is 768x576)
  // wwidth and wheight are the size of the window whatever size it is

  if (!WINDOW (video))
    return -1;

  if (!gc)
    gc = XCreateGC (DISPLAY (video), WINDOW (video), 0, NULL);

  src_w = owidth;
  src_h = oheight;

  if (hw_scaling)
    src_h = 576;

  if (!xvimage)
    {
      xvimage =
        x11_create_xvimage (DISPLAY (video), src_w, src_h,
                            &ximage_shm);
      if (NULL == xvimage)
        {
          fprintf (stderr, "oops: can't create xvimage\n");
          exit (1);
        }
    }

  if (direct_video)
    {
      if (NULL == cb (video_offset, owidth, oheight, 0, direct_xsize))
        return -1;
    }
  else
    {
      image_data = xvimage->data;
      if (cur_deinterlace != DEINTERLACE_NONE)
        {
          if (NULL == cb (deinterlace_buffer[0], owidth, /*oheight*/src_h, 0, owidth))
            return -1;
          deinterlace_yuv (image_data, deinterlace_buffer, owidth*2, src_h, cur_deinterlace);
        }
      else
        {
          if (NULL == cb (image_data, owidth, /*oheight*/src_h, 0, owidth))
            return -1;
        }

      if (plugin)
        {
          ysbar = plugin->get_ybar();

          ydbar = (int)(ysbar * (double)(wheight) / (double)maxheight[cur_norm] );

          if (!hw_scaling)
            ysbar = ydbar; // if sw scaling, then image is already scaled as the black border...
        }
      else
        {
          ysbar = 0;
          ydbar = 0;
        }

#if defined(HAVE_DIVX4LINUX) || defined (HAVE_FFMPEG) || defined(HAVE_XVID)
      if (is_divx_in_progress())
        {
          divx_encode(image_data+ydbar*owidth*2);

          if (divx_display_frames() == 0)
            return 0;
        }
#endif

      //XLockDisplay (DISPLAY (video));
      if (ydbar>0)
        {
          // add extra 2 lines to clear nasty green or rosa line noisy lines
          XClearArea (DISPLAY(video), WINDOW(video), 0, 0, wwidth, ydbar+2, False);
          XClearArea (DISPLAY(video), WINDOW(video), 0, wheight-ydbar-2, wwidth, ydbar+2, False);
        }
          

      XVPUTIMAGE (DISPLAY (video), xv_img_port, WINDOW (video), gc, xvimage,
                  //0, 0+ysbar, owidth, /*oheight*/src_h-2*ysbar, // source rect
                  //0, 0+ydbar, wwidth, wheight-2*ydbar); // dest rect
                  0, 0+ysbar, owidth, /*oheight*/src_h-2*ysbar, // source rect
                  0, 0+ydbar, wwidth, wheight-2*ydbar); // dest rect

      //XFlush (DISPLAY (video));
      //XUnlockDisplay (DISPLAY (video));
     }

  if(cmd) {do_command_mult(cmd);cmd=NULL;}
  return 0;
}
#endif

Widget
video_init (Widget parent)
{
  Window root = DefaultRootWindow (DISPLAY (parent));
  
  swidth = SCREEN (parent)->width;
  sheight = SCREEN (parent)->height;

  x11_native_format = x11_init (DISPLAY (parent), XtWindow(parent) );
  video_parent = parent;

  video = XtVaCreateManagedWidget ("tv", simpleWidgetClass, parent, NULL);
  
  /* TV Widget -- need visibility, expose */
  XtAddEventHandler (video,
                     VisibilityChangeMask |
                     StructureNotifyMask, False, video_event, NULL);

  /* Shell widget -- need map, unmap, configure */
  XtAddEventHandler (parent, StructureNotifyMask, True, video_event, NULL);

  /* root window -- need */
  XSelectInput (DISPLAY (video), root,
                VisibilityChangeMask |
                SubstructureNotifyMask | StructureNotifyMask);

  XtRegisterDrawable (DISPLAY (video), root, video);

  return video;
}

void
video_lock_size(void)
{
  /* TV Widget -- need visibility, expose */
  XtRemoveEventHandler (video,
                     StructureNotifyMask, False, video_event, NULL);

  /* Shell widget -- need map, unmap, configure */
  XtRemoveEventHandler (video_parent, StructureNotifyMask, True, video_event, NULL);
  //XtVaSetValues(tv, XtNallowShellResize, False, NULL);

}

void
video_unlock_size(void)
{
  /* TV Widget -- need visibility, expose */
  XtAddEventHandler (video,
                     StructureNotifyMask, False, video_event, NULL);

  /* Shell widget -- need map, unmap, configure */
  XtAddEventHandler (video_parent, StructureNotifyMask, True, video_event, NULL);
  //  XtVaSetValues(tv, XtNallowShellResize, True, NULL);
}

void
video_background(void)
{
  Display *display;
  int screen_num;
  Screen *screen;
  Window window;
  //XVisualInfo vinfo;
  
  XColor background, ignored;
  XSetWindowAttributes attr;
  //Colormap              theCmap = 0;

  display = DISPLAY(video);
  screen = DefaultScreenOfDisplay(display);
  screen_num = DefaultScreen(display);
  window = WINDOW(video);

  if (XAllocNamedColor (display,
                    DefaultColormap (display, screen_num),
                    "black",
                    &background,
                        &ignored) == 0)
    fprintf (stderr, "video: Cannot allocate color black\n");

  //attr.background_pixel = background.pixel;
  attr.background_pixel = BlackPixel(display, screen_num);
  attr.border_pixel     = BlackPixel(display, screen_num);
  attr.override_redirect = True;

  XChangeWindowAttributes(display,
                          window,
                          CWBackPixel | CWBorderPixel,
                          //|CWOverrideRedirect,
                          &attr);

  /*
    XMatchVisualInfo(display, screen_num, DefaultDepthOfScreen (screen), TrueColor, &vinfo);

    theCmap   = XCreateColormap(display,
                                RootWindow(display, screen_num),
                                vinfo.visual, AllocNone);

    attr.background_pixel = 0;
    attr.border_pixel     = 1;
    attr.colormap         = theCmap;
    XChangeWindowAttributes(display,
                            window,
                            CWBackPixel | CWBorderPixel | CWColormap,
                            &attr);
  */

  XClearWindow(display, window);
  XFlush(display);
}

void
video_close (void)
{
  Window root = DefaultRootWindow (DISPLAY (video));

  XSelectInput (DISPLAY (video), root, 0);
  XtUnregisterDrawable (DISPLAY (video), root);

  if (ximage)
    {
      x11_destroy_ximage (DISPLAY (video), ximage, ximage_shm);
      ximage = NULL;
    }

#ifdef HAVE_XV

  if (xvimage)
    {
      x11_destroy_xvimage (DISPLAY (video), xvimage, ximage_shm);
      xvimage = NULL;
    }

  if (have_xv)
    {

      if (XvUngrabPort (DISPLAY (video), xv_img_port, CurrentTime) != Success)
        fprintf (stderr, "XvUngrabPort() failed.\n");
    }

#ifdef HAVE_VIDEO_XV
  if (have_video_xv)
    {
      if (XvUngrabPort (DISPLAY (video), xv_video_port, CurrentTime) != Success)
        fprintf (stderr, "XvUngrabPort() failed.\n");
    }
#endif

#endif

}

void
video_get_size(int *w, int *h)
{
  *w = owidth;
  *h = oheight;
}

#ifdef HAVE_VIDEO_XV
unsigned long
xv_getfreq(void)
{
#ifdef HAVE_XV
  unsigned int freq;

  if (xv_video_freq != -1)
    XvGetPortAttribute(DISPLAY(video), xv_video_port, xv_video_freq, &freq);
  printf("xv_getfreq: freq = %d\n", freq);
  return (unsigned long)freq;
#endif
}

void
xv_setfreq(unsigned long freq)
{
#ifdef HAVE_XV
  if (xv_video_freq != -1)
    XvSetPortAttribute(DISPLAY(video), xv_video_port, xv_video_freq, freq);

  XSync(DISPLAY(video), False);
  printf("xv_setfreq: freq = %ld\n", freq);
#endif
}

int
xv_tuned(void)
{
    /* don't know ... */
    return 0;
}

void
xv_setinput(int input, int norm)
{
#ifdef HAVE_XV
  /*
if (input != -1)
  XvSetPortAttribute(DISPLAY(video), xv_video_port, xv_video_input, input);

if (input != -1)
  XvSetPortAttribute(DISPLAY(video), xv_video_port, xv_video_norm, norm);

 XSync(DISPLAY(video), False);
  */
#endif
}

void
xv_setpicture(struct video_picture *pict)
{
#ifdef HAVE_XV
  /*
  if (xv_video_colour != -1)
    XvSetPortAttribute(DISPLAY(video), xv_video_port, xv_video_colour, pict->colour);
  */
  if (xv_video_brightness != -1)
    XvSetPortAttribute(DISPLAY(video), xv_video_port, xv_video_brightness, pict->brightness);
  if (xv_video_contrast != -1)
    XvSetPortAttribute(DISPLAY(video), xv_video_port, xv_video_contrast, pict->contrast);
  /*
  if (xv_video_saturation != -1)
    XvSetPortAttribute(DISPLAY(video), xv_video_port, xv_video_saturation, pict->saturation);
  */
  if (xv_video_hue != -1)
    XvSetPortAttribute(DISPLAY(video), xv_video_port, xv_video_hue, pict->hue);

  XSync(DISPLAY(video), False);
#endif
}

void
xv_getaudio(struct video_audio *audio)
{
#ifdef HAVE_XV
  int value;

  if (xv_video_volume != -1)
    {
      XvGetPortAttribute(DISPLAY(video), xv_video_port, xv_video_volume, &value);
      audio->volume = value;
    }

  if (xv_video_mute != -1)
    {
      XvGetPortAttribute(DISPLAY(video), xv_video_port, xv_video_mute, &value);
      if (value)
        audio->flags |= VIDEO_AUDIO_MUTE;
      else
        audio->flags &= ~VIDEO_AUDIO_MUTE;
    }
#endif
}

void
xv_setaudio(struct video_audio *audio)
{
#ifdef HAVE_XV
  int value;

  if (xv_video_volume != -1)
    {
      value = (int)((float)(audio->volume)/65535.0*2000.0-1000.0);
      XvSetPortAttribute(DISPLAY(video), xv_video_port, xv_video_volume, value);
      fprintf(stderr, "xv_setaudio: v4l vol = %d, XV vol = %d\n", audio->volume, value);
    }

  if (xv_video_mute != -1)
    {
      if (audio->flags & VIDEO_AUDIO_MUTE)
        {
          XvSetPortAttribute(DISPLAY(video), xv_video_port, xv_video_mute, 1);
          fprintf(stderr, "xv_setaudio: mute\n");
        }
      else
        {
          XvSetPortAttribute(DISPLAY(video), xv_video_port, xv_video_mute, 0);
          fprintf(stderr, "xv_setaudio: unmute\n");
        }
    }
  XSync(DISPLAY(video), False);
#endif
}
#endif // HAVE_VIDEO_XV
